@@ -3,8 +3,12 @@
 
 
 from __future__ import absolute_import
-import six
+from __future__ import print_function
+import errno
 import numbers
+import os
+import six
+import sys
 
 
 class PycommandsException(Exception):
@@ -19,6 +23,10 @@ class DoesNotExist(PycommandsException):
     pass
 
 
+class InvalidInputArgument(PycommandsException):
+    pass
+
+
 class DataObjectDoesNotExist(DoesNotExist):
     pass
 
@@ -35,9 +43,11 @@ class UserDoesNotExist(DoesNotExist):
     pass
 
 
-class UserGroupDoesNotExist(DoesNotExist):
+class GroupDoesNotExist(DoesNotExist):
     pass
 
+# NOTE: Everything of the form *UserGroup* is deprecated.
+UserGroupDoesNotExist = GroupDoesNotExist
 
 class ResourceDoesNotExist(DoesNotExist):
     pass
@@ -59,16 +69,85 @@ class MultipleResultsFound(QueryException):
     pass
 
 
+class NotImplementedInIRODSServer(PycommandsException):
+    def __init__(self, feature_description, required_iRODS_version = ()):
+        super(NotImplementedInIRODSServer,self).__init__(feature_description + ': Not supported by the connected iRODS server.')
+        self.required_iRODS_version = required_iRODS_version
+    def __str__(self):
+        nv = self.required_iRODS_version
+        return '{}{}'.format(self.args, ' [requires iRODS version: {nv}]'.format(**locals()) if nv else '')
+    def __repr__(self):
+        return self.__class__.__name__ + str(self)
+
+
 class iRODSExceptionMeta(type):
     codes = {}
-
+    positive_code_error_message = "For {name}, a positive code of {attrs[code]} was declared."
     def __init__(self, name, bases, attrs):
         if 'code' in attrs:
+            if attrs['code'] > 0:
+                print(self.positive_code_error_message.format(**locals()), file = sys.stderr)
+                exit(1)
             iRODSExceptionMeta.codes[attrs['code']] = self
 
 
+class Errno:
+    """Encapsulates an integer error code from the operating system
+       and provides a text representation.
+    """
+    def __init__(self,arg0,*_):
+        """Initializes an object with an integer error code arg0.
+           Further arguments are optional and ignored.
+        """
+        self.int_code = arg0
+
+    def __repr__(self):
+        e = self.int_code
+        try:
+            return self.__class__.__name__ + repr(tuple([e, errno.errorcode[e], os.strerror(e)]))
+        except:
+            # The errno code is unrecognized, so fall through to default representation.
+            pass
+        return self.__class__.__name__ + repr(tuple([e,]))
+
+    def __int__(self):
+        return self.int_code
+
+
 class iRODSException(six.with_metaclass(iRODSExceptionMeta, Exception)):
-    pass
+    """An exception that originates from a server error.
+       Exception classes that are derived from this base and represent a concrete error, should
+       store a unique error code (X*1000) in their 'code' attribute, where X < 0.
+    """
+
+    def __init__(self,*arg):
+        explicit_errno = None
+        argl = list(arg)
+
+        # For consistency with the text representation, allow initialization with an Errno object
+        # at the end of the argument list.
+        # Example: err = UNIX_FILE_OPEN_ERR('message', Errno(13))
+        #          err_copy = eval(repr(err))
+        if hasattr(self.__class__,'code'):
+            if argl and isinstance (argl[-1],Errno):
+                explicit_errno = argl.pop()
+
+        super(iRODSException,self).__init__(*argl)
+
+        # To properly represent the Errno instance, the "code" attribute should be (-errno) plus the thousands
+        # attribute stored in the class's code attribute.  Because Python honors the instance "code" attribute
+        # over the class "code" attribute, the following does _not_ modify the class:
+        if explicit_errno:
+            self.code -= int(explicit_errno)
+
+    def __repr__(self):
+        args = tuple(self.args)
+        code = getattr(self,'code',None)
+        if code is not None:
+            os_err = abs(code) % 1000
+            if os_err: args += (Errno(os_err),)
+        return self.__class__.__name__ + repr(args)
+
 
 
 def nominal_code( the_code, THRESHOLD = 1000 ):
@@ -562,6 +641,10 @@ class SYS_REPLICA_DOES_NOT_EXIST(iRODSException):
     code = -164000
 
 
+class SYS_REPLICA_INACCESSIBLE(iRODSException):
+    code = -168000
+
+
 class SYS_NOT_ALLOWED(iRODSException):
     code = -169000
 
@@ -1271,6 +1354,10 @@ class CAT_TICKET_EXPIRED(TicketException):
     code = -891000
 
 
+class CAT_TICKET_USES_EXCEEDED(TicketException):
+    code = -892000
+
+
 class MiscException(iRODSException):
     pass
 
@@ -1587,6 +1674,10 @@ class DATAACCESSINX_EMPTY_IN_STRUCT_ERR(RuleEngineException):
     code = -1016000
 
 
+class RE_RUNTIME_ERROR(RuleEngineException):
+    code = -1205000
+
+
 class RE_TYPE_ERROR(RuleEngineException):
     code = -1230000
 
@@ -1978,6 +2069,9 @@ class PHP_REQUEST_STARTUP_ERR(PHPException):
 class PHP_OPEN_SCRIPT_FILE_ERR(PHPException):
     code = -1602000
 
+class DIRECT_CHILD_ACCESS(iRODSException):
+    code = -1816000
+
 class PAMException(iRODSException):
     pass
 
